@@ -17,21 +17,101 @@ Scripts
 arfx:      general-purpose compression/extraction utility with tar-like syntax
 """
 
-__version__ = "2.0.1"
-
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import unicode_literals
 import os
 import sys
 import posixpath
 import argparse
+import logging
+
 import arf
 from . import io
-from tools import filecache
+
+__version__ = "2.2.0"
 
 # template for extracted files
 default_extract_template = "{entry}_{channel}.wav"
 # template for created entries
 default_entry_template = "{base}_{index:04}"
 
+log = logging.getLogger('arfx')   # root logger
+
+
+def entry_repr(entry):
+    from h5py import h5t
+    attrs = entry.attrs
+    datatypes = arf.DataTypes._todict()
+    out = "%s" % (entry.name)
+    for k, v in attrs.items():
+        if k.isupper():
+            continue
+        if k == 'timestamp':
+            out += ("\n  timestamp : %s" %
+                    arf.timestamp_to_datetime(v).strftime("%Y-%m-%d %H:%M:%S.%f"))
+        elif k == 'uuid':
+            out += "\n  uuid : %s" % v
+        else:
+            out += "\n  %s : %s" % (k, v)
+    for name, dset in entry.items():
+        out += "\n  /%s :" % name
+        if isinstance(dset.id.get_type(), h5t.TypeVlenID):
+            out += " vlarray"
+        else:
+            out += " array %s" % (dset.shape,)
+            if 'sampling_rate' in dset.attrs:
+                out += " @ %.1f/s" % dset.attrs['sampling_rate']
+            if dset.dtype.names is not None:
+                out += " (compound type)"
+
+        out += ", units '%s'" % dset.attrs.get('units', '')
+        out += ", type %s" % datatypes[dset.attrs.get('datatype',
+                                                      arf.DataTypes.UNDEFINED)]
+        if dset.compression:
+            out += " [%s%d]" % (dset.compression, dset.compression_opts)
+    return out
+
+
+def dataset_properties(dset):
+    """Infers the type of data and some properties of an hdf5 dataset.
+
+    Returns tuple: (sampled|event|interval|unknown), (array|table|vlarry), ncol
+    """
+    from h5py import h5t
+    interval_dtype_names = ('name', 'start', 'stop')
+    dtype = dset.id.get_type()
+
+    if isinstance(dtype, h5t.TypeVlenID):
+        return 'event', 'vlarray', dset.id.shape[0]
+
+    if isinstance(dtype, h5t.TypeCompoundID):
+        # table types; do a check on the dtype for backwards compat with 1.0
+        names, ncol = dtype.dtype.names, dtype.get_nmembers()
+        if 'start' not in names:
+            contents = 'unknown'
+        elif any(k not in names for k in interval_dtype_names):
+            contents = 'event'
+        else:
+            contents = 'interval'
+        return contents, 'table', ncol
+
+    dtt = dset.attrs.get('datatype', 0)
+    ncols = len(dset.shape) < 2 and 1 or dset.shape[1]
+    if dtt < arf.DataTypes.EVENT:
+        # assume UNKNOWN is sampled
+        return 'sampled', 'array', ncols
+    else:
+        return 'event', 'array', ncols
+
+
+def pluralize(n, sing='', plurl='s'):
+    """Returns 'sing' if n == 1, else 'plurl'"""
+    if n == 1:
+        return sing
+    else:
+        return plurl
+
 
 def parse_name_template(node, template, index=0, default="NA"):
     """ Generates names for output files using a template and the entry/dataset attributes
@@ -52,10 +132,10 @@ def parse_name_template(node, template, index=0, default="NA"):
     values = dict()
     entry = dset = None
     if isinstance(node, Group):
-        entry = arf.entry.promote(node)
+        entry = node
     elif isinstance(node, Dataset):
         dset = node
-        entry = arf.entry.promote(dset.parent)
+        entry = dset.parent
 
     try:
         for lt, field, fs, c in f.parse(template):
@@ -63,11 +143,13 @@ def parse_name_template(node, template, index=0, default="NA"):
                 continue
             elif field == "entry":
                 if not entry:
-                    raise ValueError, "can't resolve {entry} field for %s" % node
+                    raise ValueError(
+                        "can't resolve {entry} field for %s" % node)
                 values[field] = pp.basename(entry.name)
             elif field == "channel":
                 if not dset:
-                    raise ValueError, "can't resolve {channel} field for %s" % node
+                    raise ValueError(
+                        "can't resolve {channel} field for %s" % node)
                 values[field] = pp.basename(dset.name)
             elif field == "index":
                 values[field] = index
@@ -85,22 +167,22 @@ def parse_name_template(node, template, index=0, default="NA"):
             return f.format(template, **values)
         else:
             return template  # no substitutions were made
-    except ValueError, e:
+    except ValueError as e:
         raise ValueError("template error: " + e.message)
 
 
 def iter_entries(src, cbase='pcm'):
-    """
-    Iterate through the entries and channels of a data source.
+    """Iterate through the entries and channels of a data source.
+
     Yields (data, entry index, entry name,)
     """
     fp = io.open(src, 'r')
     fbase = os.path.splitext(os.path.basename(src))[0]
     nentries = getattr(fp, 'nentries', 1)
-    for entry in xrange(nentries):
+    for entry in range(nentries):
         try:
             fp.entry = entry
-        except:
+        except AttributeError:
             pass
 
         if nentries == 1:
@@ -120,6 +202,7 @@ def add_entries(tgt, files, **options):
     Additional keyword arguments specify metadata on the newly created
     entries.
     """
+    from h5py import Group
     compress = options.get('compress', None)
     ebase = options.get('template', None)
     metadata = options.get('attrs', None) or dict()
@@ -127,40 +210,43 @@ def add_entries(tgt, files, **options):
     chan = "pcm"                # only pcm data can be imported
 
     if len(files) == 0:
-        raise ValueError, "must specify one or more input files"
+        raise ValueError("must specify one or more input files")
 
-    with arf.file(tgt, 'a', strict_version=True) as arfp:
+    with arf.open_file(tgt, 'a') as arfp:
+        arf.check_file_version(arfp)
+        arf.set_attributes(
+            arfp, file_creator="org.meliza.arfx/arfx " + __version__,
+            overwrite=False)
         for f in files:
-            try:
-                for fp, entry_index, entry_name in iter_entries(f):
-                    timestamp = getattr(fp, 'timestamp', None)
-                    if timestamp is None:
-                        # kludge for ewave
-                        if hasattr(fp, 'fp') and hasattr(fp.fp, 'fileno'):
-                            timestamp = os.fstat(fp.fp.fileno()).st_mtime
-                        else:
-                            raise ValueError, "%s/%d missing required timestamp" % (
-                                f, entry_index)
-                    if not hasattr(fp, 'sampling_rate'):
-                        raise ValueError, "%s/%d missing required sampling_rate attribute" % (
-                            f, entry_index)
+            for fp, entry_index, entry_name in iter_entries(f):
+                timestamp = getattr(fp, 'timestamp', None)
+                if timestamp is None:
+                    # kludge for ewave
+                    if hasattr(fp, 'fp') and hasattr(fp.fp, 'fileno'):
+                        timestamp = os.fstat(fp.fp.fileno()).st_mtime
+                    else:
+                        raise ValueError("%s/%d missing required timestamp" %
+                                         (f, entry_index))
+                if not hasattr(fp, 'sampling_rate'):
+                    raise ValueError(
+                        "%s/%d missing required sampling_rate attribute" %
+                        (f, entry_index))
 
-                    if ebase is not None:
-                        entry_name = default_entry_template.format(
-                            base=ebase, index=arfp.nentries)
-                    entry = arfp.create_entry(
-                        entry_name, timestamp, **metadata)
-
-                    entry.add_data(chan, fp.read(),
+                if ebase is not None:
+                    entry_name = default_entry_template.format(
+                        base=ebase,
+                        index=arf.count_children(arfp, Group))
+                entry = arf.create_entry(arfp, entry_name, timestamp,
+                                         entry_creator="org.meliza.arfx/arfx " +
+                                         __version__,
+                                         **metadata)
+                arf.create_dataset(entry, chan, fp.read(),
                                    datatype=datatype,
                                    sampling_rate=fp.sampling_rate,
                                    compression=compress,
                                    source_file=f,
                                    source_entry=entry_index)
-                    if options['verbose']:
-                        print "%s/%d -> /%s/%s" % (f, entry_index, entry_name, chan)
-            except Exception, e:
-                print "arfx: input error: %s" % e
+                log.debug("%s/%d -> /%s/%s", f, entry_index, entry_name, chan)
 
 
 def create_and_add_entries(tgt, files, **options):
@@ -181,15 +267,21 @@ def extract_entries(src, entries, **options):
     entry_base: if specified, name the output files sequentially
     """
     if not os.path.exists(src):
-        raise IOError, "the file %s does not exist" % src
+        raise IOError("the file %s does not exist" % src)
 
     if len(entries) == 0:
         entries = None
     ebase = options.get('template', None)
 
-    with arf.file(src, 'r', strict_version=True) as arfp:
-        for index, (ename, entry) in enumerate(arfp.items(key='timestamp')):
+    with arf.open_file(src, 'r') as arfp:
+        try:
+            arf.check_file_version(arfp)
+        except Warning as e:
+            log.warn("warning: %s", e)
+        for index, ename in enumerate(arf.keys_by_creation(arfp)):
+            entry = arfp[ename]
             attrs = dict(entry.attrs)
+            mtime = attrs.get('timestamp', [None])[0]
             if entries is None or ename in entries:
                 for channel in entry:
                     dset = entry[channel]
@@ -200,22 +292,17 @@ def extract_entries(src, entries, **options):
                     fname = parse_name_template(dset,
                                                 ebase or default_extract_template,
                                                 index=index)
-                    fp = io.open(fname, 'w', **attrs)
-                                 # will throw error for unsupported format
-
-                    dtype, stype, ncols = arf.dataset_properties(dset)
+                    dtype, stype, ncols = dataset_properties(dset)
                     if dtype != 'sampled':
+                        log.debug(
+                            "%s -> skipped (no supported containers)", dset.name)
                         continue
 
-                    fp.write(dset)
-                    if 'timestamp' in entry.attrs:
-                        fp.timestamp = entry.attrs['timestamp'][0]
-                    else:
-                        # try to set the modification time
-                        pass
+                    with io.open(fname, 'w', **attrs) as fp:
+                        fp.write(dset)
+                    os.utime(fname, (os.stat(fname).st_atime, mtime))
 
-                    if options['verbose']:
-                        print "%s -> %s" % (dset.name, fname)
+                    log.debug("%s -> %s", dset.name, fname)
 
 
 def delete_entries(src, entries, **options):
@@ -226,23 +313,23 @@ def delete_entries(src, entries, **options):
     repack: if True (default), repack the file afterward to reclaim space
     """
     if not os.path.exists(src):
-        raise IOError, "the file %s does not exist" % src
+        raise IOError("the file %s does not exist" % src)
     if entries is None or len(entries) == 0:
         return
 
-    with arf.file(src, 'r+', strict_version=True) as arfp:
+    with arf.open_file(src, 'r+') as arfp:
+        arf.check_file_version(arfp)
         count = 0
         for entry in entries:
             if entry in arfp:
                 try:
                     del arfp[entry]
                     count += 1
-                    if options['verbose']:
-                        print "deleted /%s" % entry
-                except Exception, e:
-                    print "Error deleting %s: %s" % (entry, e)
-            elif options['verbose']:
-                print "/%s: no such entry" % entry
+                    log.debug("deleted /%s", entry)
+                except Exception as e:
+                    log.error("unable to delete %s: %s", entry, e)
+            else:
+                log.debug("unable to delete %s: no such entry", entry)
     if count > 0 and options['repack']:
         repack_file(src, **options)
 
@@ -255,10 +342,13 @@ def copy_entries(tgt, files, **options):
 
     entry_base: if specified, rename entries sequentially in target file
     """
+    from .tools import memoized
+    from h5py import Group
     ebase = options.get('template', None)
-    acache = filecache(arf.file)
+    acache = memoized(arf.open_file)
 
-    with arf.file(tgt, 'a', strict_version=True) as arfp:
+    with arf.open_file(tgt, 'a') as arfp:
+        arf.check_file_version(arfp)
         for f in files:
             # this is a bit tricky:
             # file.arf is a file; file.arf/entry is entry
@@ -266,32 +356,26 @@ def copy_entries(tgt, files, **options):
             # on windows, dir\file.arf/entry is an entry
             pn, fn = posixpath.split(f)
             if os.path.isfile(f):
-                it = ((f, entry) for ename, entry in acache[f].items())
+                it = ((f, entry) for ename, entry in acache(f).items())
             elif os.path.isfile(pn):
-                fp = acache[pn]
+                fp = acache(pn)
                 if fn in fp:
                     it = ((pn, fp[fn]),)
                 else:
-                    print "Error: no such entry %s" % f
+                    log.error("unable to copy %s: no such entry", f)
                     continue
             else:
-                print "Error: %s does not exist" % f
+                log.error("unable to copy %s: does not exist", f)
                 continue
 
             for fname, entry in it:
-                try:
-                    if ebase is not None:
-                        entry_name = default_entry_template.format(
-                            base=ebase, index=arfp.nentries)
-                    else:
-                        entry_name = posixpath.basename(entry.name)
-                    arfp.h5.copy(entry, arfp.h5, name=entry_name)
-                    if options['verbose']:
-                        print "%s%s -> %s/%s" % (fname, entry.name, tgt, entry_name)
-                except Exception, e:
-                    print "arfx: copy error for %s%s: %s" % (fname, entry.name, e)
-
-    acache.__exit__(None, None, None)
+                if ebase is not None:
+                    entry_name = default_entry_template.format(
+                        base=ebase, index=arf.count_children(arfp, Group))
+                else:
+                    entry_name = posixpath.basename(entry.name)
+                arfp.copy(entry, arfp, name=entry_name)
+                log.debug("%s%s -> %s/%s", fname, entry.name, tgt, entry_name)
 
 
 def list_entries(src, entries, **options):
@@ -302,20 +386,29 @@ def list_entries(src, entries, **options):
              that are in this list (more verbosely)
     """
     if not os.path.exists(src):
-        raise IOError, "the file %s does not exist" % src
-    print "%s:" % src
-    with arf.file(src, 'r', strict_version=True) as arfp:
+        raise IOError("the file %s does not exist" % src)
+    print("%s:" % src)
+    with arf.open_file(src, 'r') as arfp:
+        try:
+            arf.check_file_version(arfp)
+        except Warning as e:
+            log.warn("warning: %s", e)
         if entries is None or len(entries) == 0:
-            for name, entry in arfp.items(key='timestamp'):
+            try:
+                it = arf.keys_by_creation(arfp)
+            except RuntimeError:
+                it = iter(arfp)
+            for name in it:
+                entry = arfp[name]
                 if options.get('verbose', False):
-                    print entry
+                    print(entry_repr(entry))
                 else:
-                    print "%s: %d channel%s" % (entry.name, entry.nchannels,
-                                                arf.pluralize(entry.nchannels))
+                    print("%s: %d channel%s" %
+                          (entry.name, len(entry), pluralize(len(entry))))
         else:
             for ename in entries:
                 if ename in arfp:
-                    print arfp[ename]
+                    print(entry_repr(arfp[ename]))
 
 
 def update_entries(src, entries, **options):
@@ -326,31 +419,36 @@ def update_entries(src, entries, **options):
              name parameter is set, the entries are renamed sequentially
     """
     if not os.path.exists(src):
-        raise IOError, "the file %s does not exist" % src
+        raise IOError("the file %s does not exist" % src)
     ebase = options.get('template', None)
     if (entries is None or len(entries) == 0) and ebase is not None:
         if ebase.find('{') < 0:
-            raise ValueError, "with multiple entries, template needs to have {} formatter fields"
+            raise ValueError(
+                "with multiple entries, template needs to have {} formatter fields")
     metadata = options.get('attrs', None) or dict()
     if 'datatype' in options:
         metadata['datatype'] = options['datatype']
 
-    with arf.file(src, 'r+', strict_version=True) as arfp:
+    with arf.open_file(src, 'r+') as arfp:
+        try:
+            arf.check_file_version(arfp)
+        except Warning as e:
+            log.warn("warning: %s", e)
         for i, entry in enumerate(arfp):
             if entries is None or len(entries) == 0 or posixpath.relpath(entry) in entries:
                 enode = arfp[entry]
                 if options.get('verbose', False):
-                    print "vvvvvvvvvv"
-                    print enode.__str__()
-                    print "**********"
+                    print("vvvvvvvvvv")
+                    print(entry_repr(enode))
+                    print("**********")
                 if ebase:
                     name = parse_name_template(enode, ebase, index=i)
-                    arfp.h5[name] = enode
-                    del arfp.h5[entry]  # entry object should remain valid
-                arfp.set_attributes(enode, **metadata)
+                    arfp[name] = enode
+                    del arfp[entry]  # entry object should remain valid
+                arf.set_attributes(enode, **metadata)
                 if options.get('verbose', False):
-                    print enode.__str__()
-                    print "^^^^^^^^^^"
+                    print(entry_repr(enode))
+                    print("^^^^^^^^^^")
 
 
 def repack_file(path, **options):
@@ -364,14 +462,10 @@ def repack_file(path, **options):
         cmd += "-f SHUF -f GZIP=%d " % compress
     try:
         tdir = mkdtemp()
-        if options['verbose']:
-            sys.stdout.write("Repacking %s..." % path)
-            sys.stdout.flush()
+        log.info("Repacking %s", path)
         fdir, fbase = os.path.split(path)
         os.system(cmd + path + " " + os.path.join(tdir, fbase))
         copy(os.path.join(tdir, fbase), path)
-        if options['verbose']:
-            sys.stdout.write("done\n")
     finally:
         rmtree(tdir)
 
@@ -389,7 +483,8 @@ class ParseKeyVal(argparse.Action):
         if kv is None:
             kv = dict()
         if not arg.count('=') == 1:
-            print >> sys.stderr, "-k %s argument badly formed; needs key=value" % arg
+            raise ValueError(
+                "-k %s argument badly formed; needs key=value" % arg)
         else:
             key, val = arg.split('=')
             kv[key] = val
@@ -402,20 +497,21 @@ class ParseDataType(argparse.Action):
         if not arg.isdigit():
             arg = arf.DataTypes._fromstring(arg)
             if arg is None:
-                raise ValueError, "%s is not a valid data type" % arg
+                raise ValueError("%s is not a valid data type" % arg)
         setattr(namespace, self.dest, int(arg))
 
 
 def arfx():
+    import datetime
 
     p = argparse.ArgumentParser(
         description='copy data in and out of ARF files')
     p.add_argument('entries', nargs='*')
     p.add_argument('--version', action='version',
                    version='%(prog)s ' + __version__)
-    p.add_argument(
-        '--help-datatypes', help='print available datatypes and exit',
-        action='version', version=arf.DataTypes._doc())
+    p.add_argument('--help-datatypes',
+                   help='print available datatypes and exit',
+                   action='version', version=arf.DataTypes._doc())
 
     # operations
     pp = p.add_argument_group('Operations')
@@ -449,24 +545,39 @@ def arfx():
                    default=arf.DataTypes.UNDEFINED, metavar='DATATYPE', dest='datatype', action=ParseDataType)
     g.add_argument(
         '-k', help='specify attributes of entries', action=ParseKeyVal,
-                   metavar="KEY=VALUE", dest='attrs')
+        metavar="KEY=VALUE", dest='attrs')
     g.add_argument(
         '-P', help="don't repack when deleting entries", action='store_false',
-                   dest='repack')
+        dest='repack')
     g.add_argument(
         '-z', help="set compression level in ARF (default: %(default)s)", type=int,
-                   default=1, dest='compress')
+        default=1, dest='compress')
 
     args = p.parse_args()
 
+    ch = logging.StreamHandler()
+    formatter = logging.Formatter("[%(name)s] %(message)s")
+    if args.verbose:
+        loglevel = logging.DEBUG
+    else:
+        loglevel = logging.INFO
+    log.setLevel(loglevel)
+    ch.setLevel(loglevel)  # change
+    ch.setFormatter(formatter)
+    log.addHandler(ch)
+    log.info("version: %s", __version__)
+    log.info("run time: %s", datetime.datetime.now())
+
     try:
         opts = args.__dict__.copy()
         entries = opts.pop('entries')
         args.op(args.arffile, entries, **opts)
-    except Exception, e:
-        print >> sys.stderr, "arfx: error: %s" % e
+    except Exception as e:
+        raise
+        print("arfx: error: %s" % e)
         if isinstance(e, DeprecationWarning):
-            print >> sys.stderr, "      use arfx --upgrade to convert to version %s" % arf.spec_version
+            print("      use arfx --upgrade to convert to version %s" %
+                  arf.spec_version)
         sys.exit(-1)
     return 0
 
